import argparse

from imagetitler.constants import *


def parse_input() -> argparse.Namespace:
    """
    Parses the command line input.

    :return: the processed command line arguments
    """
    parser = argparse.ArgumentParser()
    _add_title_option(parser)
    _add_path_option(parser)
    _add_output_path_option(parser)
    _add_tier_option(parser)
    _add_logo_path_option(parser)
    _add_batch_option(parser)
    _add_font_option(parser)
    _add_custom_size_option(parser)
    args = parser.parse_args()
    return args


def _add_title_option(parser: argparse.ArgumentParser) -> None:
    """
    A helper function which sets up the title settings for the parser.
    The title is then used to add a custom title to the image.

    :param parser: an argument parser
    :return: None
    """
    parser.add_argument(
        '-t',
        f'--{KEY_TITLE}',
        help="add a custom title to the image"
    )


def _add_path_option(parser: argparse.ArgumentParser) -> None:
    """
    A helper function which sets up the path settings for the parser.
    The path setting determines which file or folder is to be processed.

    :param parser: an argument parser
    :return: None
    """
    parser.add_argument(
        '-p',
        f'--{KEY_PATH}',
        help="select an image file (or folder when batch processing)"
    )


def _add_output_path_option(parser: argparse.ArgumentParser) -> None:
    """
    A helper function which sets up the output path settings for the parser.
    The output path setting determines where the file will be stored.

    :param parser: an argument parser
    :return: None
    """
    parser.add_argument(
        '-o',
        f'--{KEY_OUTPUT_PATH}',
        help="select an output path for the processed image"
    )


def _add_tier_option(parser: argparse.ArgumentParser) -> None:
    """
    A helper function which sets up the tier setting for the parser.
    The tier setting determines the border color of the title bars.

    :param parser: an argument parser
    :return: None
    """
    parser.add_argument(
        '-r',
        f'--{KEY_TIER}',
        choices=TIER_MAP.keys(),
        help="select an image tier"
    )


def _add_logo_path_option(parser: argparse.ArgumentParser) -> None:
    """
    A helper function which sets up the logo path setting for the parser.
    The logo path setting is used to display a logo on the image.

    :param parser: an argument parser
    :return: None
    """
    parser.add_argument(
        '-l',
        f'--{KEY_LOGO_PATH}',
        help="select a logo file for addition to the processed image"
    )


def _add_batch_option(parser: argparse.ArgumentParser) -> None:
    """
    A helper function which sets up the batch settings for the parser.
    The batch is then used to determine the type of image processing performed.

    :param parser: an argument parser
    :return: None
    """
    parser.add_argument(
        '-b',
        f'--{KEY_BATCH}',
        default=DEFAULT_BATCH_MODE,
        action='store_true',
        help="turn on batch processing"
    )


def _add_font_option(parser: argparse.ArgumentParser) -> None:
    """
    A helper function which sets up the font settings for the parser.
    The font is then used to change the appearance of the title.

    :param parser: an argument parser
    :return: None
    """
    parser.add_argument(
        '-f',
        f'--{KEY_FONT}',
        default=DEFAULT_FONT,
        help="change the default font by path (e.g. 'arial.ttf')"
    )


def _add_custom_size_option(parser: argparse.ArgumentParser) -> None:
    """
    A helper function which sets up the size options for the parser.
    The size is then used to resize and crop input images.

    :param parser: an argument parser
    :return: None
    """
    parser.add_argument(
        "-s",
        f'--{KEY_SIZE}',
        choices=SIZE_MAP.keys(),  # [f'{k} {v}' for k, v in SIZE_MAP.items()]
        help="change the default size of the output image"
    )



from pathlib import Path
from typing import List

import pathvalidate
import piexif
import piexif.helper
import pkg_resources
from PIL import Image

from imagetitler.constants import *


def save_copies(edited_images: List[Image.Image], **kwargs) -> List[str]:
    """
    Saves a list of Pillow images as image files. The typical list of options
    apply and determine what the output file name will look like. For example,
    the title keyword will apply that title to all images in the list. Meanwhile,
    if the title keyword is not present but the filename attribute of edited_image
    is present, then that attribute will be used. Otherwise, a generic file name
    is given (image-titler).

    Note: the batch setting must be present to process more than one image.

    Currently, image files are given the following name format:

    {title}-featured-image-{software version}.{extension}

    :param edited_images: a list of edited images
    :param kwargs: a set of keyword arguments (see parse_input for options)
    :return: a list of storage paths
    """
    storage_paths = list()
    if not kwargs.get(KEY_BATCH):  # batch must be turned on to process multiple images
        edited_images = edited_images[:1]
    for index, edited_image in enumerate(edited_images):
        storage_path = _generate_image_output_path(edited_image, index, **kwargs)
        exif = _generate_version_exif(edited_image)
        edited_image.save(storage_path, subsampling=0, quality=100, exif=exif)
        storage_paths.append(storage_path)
    return storage_paths


def _generate_version_exif(image: Image.Image) -> bytes:
    """
    Given an image and version, this function will place that vision in the EXIF data of the file.

    Currently, this function is limited to files that already have EXIF data. Naturally, not
    all files have EXIF data, so I'm not sure how useful this feature is. That said, titler's
    a nice start!

    :param image: an image file
    :return: the exif data as a byte string (empty string for images that didn't already have data)
    """
    exif_data = b""
    version: str = pkg_resources.require("image-titler")[0].version
    version = version.replace(".", SEPARATOR)
    if exif := image.info.get('exif'):
        exif_dict = piexif.load(exif)
        exif_dict['Exif'][piexif.ExifIFD.UserComment] = piexif.helper.UserComment.dump(f'image-titler-v{version}')
        exif_data = piexif.dump(exif_dict)
    return exif_data


def _get_file_name(edited_image: Image.Image, **kwargs) -> str:
    """
    Retrieve the underlying file name from the edited_image or
    the set of kwargs. Priority goes as follows:

    1. title option
    2. original file name
    3. path option
    4. fallback (e.g. image-titler)

    :param edited_image: the edited image
    :param kwargs: the set of options
    :return: the file name without the extension (e.g. image-titler)
    """
    if title := kwargs.get(KEY_TITLE):
        file_name = pathvalidate.sanitize_filename(title.lower().replace(" ", SEPARATOR))
    elif hasattr(edited_image, 'filename'):
        file_name = Path(edited_image.filename).stem
    elif path := kwargs.get(KEY_PATH):
        file_name = Path(path).stem if Path(path).is_file() else "image-titler"
    else:
        file_name = "image-titler"
    return file_name


def _get_extension(edited_image: Image.Image) -> str:
    """
    Gets the extension for the new image.

    :param edited_image: the edited image
    :return: the file extension with the dot (e.g. ".jpg")
    """
    extension = ".jpg"
    if hasattr(edited_image, 'filename'):
        extension = Path(edited_image.filename).suffix
    return extension


def _get_index(index: int, **kwargs) -> str:
    """
    Gets the index of the output path given the index and some options.
    This returns "-i{index}" iff the batch and title options are truthy.
    Otherwise, it generates an empty string.

    :param index: the current index of the image
    :param kwargs: a set of options
    :return: a string in the form of a file name tag (e.g. -i17) or an empty string
    """
    i = ""
    if kwargs.get(KEY_BATCH) and kwargs.get(KEY_TITLE):
        i = f'-i{index}'
    return i


def _get_version() -> str:
    """
    Gets the version of the image-titler and places in a string as follows: -v{version}

    :return: the package version as a string (e.g. -v2.0.1)
    """
    version = pkg_resources.require("image-titler")[0].version
    version = version.replace(".", SEPARATOR)
    return f'-v{version}'


def _get_output_path(**kwargs) -> str:
    """
    Gets the output path option if it exists. Otherwise, returns an empty string.

    :param kwargs: a set of options
    :return: the output path or an emtpy string
    """
    if not (output_path := kwargs.get("output_path")):
        output_path = ""
    if output_path:
        output_path += "/"
    return output_path


def _generate_image_output_path(edited_image: Image.Image, index: int, **kwargs) -> str:
    """
    A helper function which generates an image output path from an image and its options.
    If a title exists, this method will use the title as the file name.
    If the image has the filename attribute, that will be used instead.
    Otherwise, a generic file name is created.

    :param edited_image: an image to be stored
    :param index: the index of this image in a set
    :return: the path of the file to be created
    """
    version: str = _get_version()
    file_name = _get_file_name(edited_image, **kwargs)
    extension = _get_extension(edited_image)
    index = _get_index(index, **kwargs)
    output_path = _get_output_path(**kwargs)
    storage_path = f'{output_path}{file_name}{version}{index}{extension}'
    return storage_path



"""
The commandline interface for the image-titler script.
"""

from imagetitler.draw import process_images
from imagetitler.parse import parse_input
from imagetitler.store import save_copies


def main() -> None:
    """
    The main function.

    :return: None
    """
    args = vars(parse_input())
    images = process_images(**args)
    save_copies(images, **args)


if __name__ == '__main__':
    main()



import os

KEY_BATCH = "batch"
KEY_FONT = "font"
KEY_LOGO_PATH = "logo_path"
KEY_PATH = "path"
KEY_TIER = "tier"
KEY_TITLE = "title"
KEY_OUTPUT_PATH = "output_path"
KEY_SIZE = "size"

FILE_TYPES = [('image files', ('.png', '.jpg', '.jpeg'))]

SEPARATOR = "-"

DEFAULT_BATCH_MODE = False
DEFAULT_FONT = os.path.join(os.path.dirname(__file__), "assets/fonts/BERNHC.TTF")
DEFAULT_SIZE = "WordPress"

GOLD = (255, 215, 0)
SILVER = (211, 211, 211)

TIER_MAP = {
    "free": SILVER,
    "premium": GOLD
}

SIZE_MAP = {
    "Twitter": (1200, 628),  # Card size according to: https://louisem.com/217438/twitter-image-size
    "WordPress": (1200, 628),  # Thumbnail size according to: https://blog.snappa.com/youtube-thumbnail-size/
    "YouTube": (1280, 720)  # Featured image size according to: https://blog.snappa.com/wordpress-featured-image-size/
}

TRC_ICON = os.path.join(os.path.dirname(__file__), 'assets/icons/the-renegade-coder-sample-icon.png')
TRC_IMAGE = os.path.join(os.path.dirname(__file__), 'assets/images/welcome-to-the-image-titler-by-the-renegade-coder.jpg')
TRC_IMAGES = os.path.join(os.path.dirname(__file__), 'assets/images/')



"""
The functional backend to the image-titler script.
"""
from pathlib import Path
from typing import Optional, List

from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from titlecase import titlecase

from imagetitler.constants import *

TEXT_FILL = (255, 255, 255)
RECTANGLE_FILL = (201, 2, 41)
WHITE = (255, 255, 255, 0)

TOP_RECTANGLE_Y = 20
X_OFFSET = TOP_RECTANGLE_Y
LOGO_PADDING = TOP_RECTANGLE_Y


def process_images(**kwargs) -> List[Image.Image]:
    """
    The main entry point for any image editing. This function
    will never return an empty list. If no settings are provided,
    this function will return a default image with a default title.

    :return: None
    """
    is_batch: bool = kwargs.get(KEY_BATCH)
    images = list()
    if is_batch:
        kwargs[KEY_PATH] = kwargs.get(KEY_PATH) if kwargs.get(KEY_PATH) else TRC_IMAGES
        images = _process_batch(**kwargs)
    else:
        kwargs[KEY_PATH] = kwargs.get(KEY_PATH) if kwargs.get(KEY_PATH) else TRC_IMAGE
        kwargs[KEY_TITLE] = kwargs.get(KEY_TITLE) if kwargs.get(KEY_TITLE) else _convert_file_name_to_title(**kwargs)
        images.append(_process_image(**kwargs))
    return images


def _process_batch(**kwargs) -> List[Image.Image]:
    """
    Processes a batch of images.

    :pre: kwargs.get(KEY_PATH) != None
    :return: None
    """
    edited_images = list()
    input_path = kwargs.get(KEY_PATH)
    for path in os.listdir(input_path):
        absolute_path = os.path.join(input_path, path)
        image_kwargs = kwargs.copy()
        image_kwargs[KEY_PATH] = absolute_path
        image_kwargs[KEY_TITLE] = kwargs.get(KEY_TITLE) if kwargs.get(KEY_TITLE) else _convert_file_name_to_title(**image_kwargs)
        edited_image = _process_image(**image_kwargs)
        edited_images.append(edited_image)
    return edited_images


def _process_image(**kwargs) -> Image.Image:
    """
    Processes a single image.

    :pre: kwargs.get(KEY_PATH) != None and kwargs.get(KEY_TITLE) != None
    :return: the edited image or None
    """
    input_path = kwargs.get(KEY_PATH)  # TODO: might be able to speed things up by caching this
    img: Image.Image = Image.open(input_path)
    cropped_img: Image.Image = _resize_image(img, **kwargs)
    if hasattr(img, "filename"):
        cropped_img.filename = img.filename  # Ensures filename data is transferred to updated copy
    color = RECTANGLE_FILL
    if logo_path := kwargs.get(KEY_LOGO_PATH):
        logo: Image.Image = Image.open(logo_path)
        color = _get_best_top_color(logo)
        _draw_logo(cropped_img, logo, **kwargs)
    edited_image = _draw_overlay(
        cropped_img,
        color,
        **kwargs
    )
    return edited_image


def _resize_image(img: Image.Image, **kwargs) -> Image.Image:
    """
    A helper function which resizes an image. First, the image is constrained
    by its width. Then, any excess height is cropped until the desired aspect
    ratio is achieved. See "size" option for more details.

    :param img: an image to be resized
    :param kwargs: a set of options
    :return: a resized image
    """
    size = _retrieve_size_from_options(**kwargs)
    img.thumbnail((size[0], img.size[1]))
    cropped_img: Image = img.crop((0, 0, size[0], size[1]))
    return cropped_img


def _retrieve_size_from_options(**kwargs) -> tuple:
    """
    A helper function for retrieving a size tuple from the SIZE_MAP.

    :param kwargs: a set of options
    :return: a size tuple
    """
    size = SIZE_MAP.get(DEFAULT_SIZE)
    if size_key := kwargs.get(KEY_SIZE):
        size = SIZE_MAP.get(size_key)
    return size


def _convert_file_name_to_title(**kwargs) -> Optional[str]:
    """
    A helper method which converts file names into titles. If the necessary arguments aren't supplied,
    this function returns None.

    :return: a title string or None
    """
    title: Optional[str] = kwargs.get(KEY_TITLE)
    path: Optional[str] = kwargs.get(KEY_PATH)
    if not title and path:
        file_path = Path(path).resolve().stem
        title = titlecase(file_path.replace(kwargs.get("separator", SEPARATOR), ' '))
    return title


def _get_bar_height(**kwargs) -> int:
    """
    A helper function which retrieves the bar height based on the size of the image.

    :param kwargs: a set of options
    :return: the bar height in pixels
    """
    image_size = _retrieve_size_from_options(**kwargs)
    return image_size[1] // 7


def _draw_rectangle(
        draw: ImageDraw,
        position: int,
        width: int,
        color: tuple = RECTANGLE_FILL,
        **kwargs
):
    """
    Draws a rectangle over the image given a ImageDraw object and the intended
    position, width, and tier.

    :param color: the color of the overlay bar
    :param draw: an picture we're editing
    :param position: the position of the rectangle to be added
    :param width: the width of the rectangle to be added
    :return: nothing
    """
    image_width = _retrieve_size_from_options(**kwargs)[0]
    draw.rectangle(
        (
            (image_width - width - X_OFFSET * 2, position),
            (image_width, position + _get_bar_height(**kwargs))
        ),
        fill=color,
        outline=TIER_MAP.get(kwargs.get(KEY_TIER), None),
        width=4
    )


def _draw_text(draw: ImageDraw, position: tuple, text: str, font: ImageFont):
    """
    Draws text on the image.

    :param draw: the picture to edit
    :param position: the position of the text as an (x, y) tuple
    :param text: the text
    :param font: the font of the text
    :return: nothing
    """
    draw.text(
        position,
        text,
        fill=TEXT_FILL,
        font=font
    )


def _get_text_position(text_width, text_height, text_ascent, y_offset, **kwargs) -> tuple:
    """
    A helper function which places the text safely within the title block.

    A lot of work went into making sure this function behaved properly.

    :param text_width: the width of the text bounding box
    :param text_height: the height of the text without the ascent
    :param text_ascent: the height of the ascent
    :param y_offset: the y location of the title block
    :return: a tuple containing the x, y pixel coordinates of the text
    """
    image_width = _retrieve_size_from_options(**kwargs)[0]
    return (
        image_width - text_width - X_OFFSET,
        y_offset - text_ascent + (_get_bar_height(**kwargs) - text_height) / 2
    )


def _get_text_metrics(text: str, font: ImageFont):
    """
    Returns some useful metrics about the font.

    :param text: the text to be displayed
    :param font: the font object
    :return: a tuple consisting of four sections of the text (top offset, text, bottom offset)
    """
    ascent, descent = font.getmetrics()
    (width, _), (_, offset_y) = font.font.getsize(text)
    return width, offset_y, ascent - offset_y, descent


def _get_appropriate_font_size(**kwargs) -> ImageFont:
    """
    A helper function which computes the font size given a set of options.

    :param kwargs: a set of options
    :return: a font of the appropriate size
    """
    bar_height = _get_bar_height(**kwargs)
    font = kwargs.get(KEY_FONT, DEFAULT_FONT)
    font = font if font else DEFAULT_FONT
    title = kwargs.get(KEY_TITLE)
    font_size = 12
    while ImageFont.truetype(font, font_size).getsize(title)[1] < bar_height - 10:
        font_size += 1
    return ImageFont.truetype(font, font_size)


def _draw_overlay(image: Image.Image, color: tuple, **kwargs) -> Image:
    """
    Draws text over an image.

    :param image: an image
    :return: the updated image
    """
    draw = ImageDraw.Draw(image)
    font = _get_appropriate_font_size(**kwargs)

    if title := kwargs.get(KEY_TITLE):
        # Detect space (precondition for split)
        if len(title.split()) > 1:
            top_half_text, bottom_half_text = _split_string_by_nearest_middle_space(title)
        else:
            top_half_text, bottom_half_text = title, None

        # Draw top
        width, top_offset, height, _ = _get_text_metrics(top_half_text, font)
        top_position = _get_text_position(width, height, top_offset, TOP_RECTANGLE_Y, **kwargs)
        _draw_rectangle(draw, TOP_RECTANGLE_Y, width, color, **kwargs)
        _draw_text(draw, top_position, top_half_text, font)

        bottom_rectangle_y = TOP_RECTANGLE_Y + _get_bar_height(**kwargs) + TOP_RECTANGLE_Y

        # Draw bottom
        if bottom_half_text:
            width, top_offset, height, _ = _get_text_metrics(bottom_half_text, font)
            bottom_position = _get_text_position(width, height, top_offset, bottom_rectangle_y, **kwargs)
            _draw_rectangle(draw, bottom_rectangle_y, width, color, **kwargs)
            _draw_text(draw, bottom_position, bottom_half_text, font)

    return image


def _get_logo_size(**kwargs) -> tuple:
    """
    A helper function that retrieves the size of the logo based on the size of the bars.

    :param kwargs: a set of options
    :return: a logo size tuple
    """
    bar_height = _get_bar_height(**kwargs)
    return bar_height, bar_height


def _draw_logo(img: Image.Image, logo: Image.Image, **kwargs):
    """
    Adds a logo to the image if a path is provided.

    :param img: an image to be modified
    :param logo: the logo file to be added
    :return: nothing
    """
    logo_size = _get_logo_size(**kwargs)
    logo.thumbnail(logo_size)
    _, height = img.size
    img.paste(logo, (LOGO_PADDING, height - logo_size[1] - LOGO_PADDING), logo)


def _split_string_by_nearest_middle_space(input_string: str) -> tuple:
    """
    Splits a string by the nearest middle space. Assumes space is in string.

    :param input_string: some string
    :return: a pair of strings
    """
    index = len(input_string) // 2
    curr_char = input_string[index]
    count = 1
    while not curr_char.isspace():
        index += (-1) ** (count + 1) * count  # thanks wolfram alpha (1, -2, 3, -4, ...)
        curr_char = input_string[index]
        count += 1
    return input_string[:index], input_string[index + 1:]


def _get_best_top_color(image: Image.Image) -> tuple:
    """
    Computes the most popular non-white color from an image.

    :param image: an image file
    :return: the most dominant color as a tuple
    """
    top_colors = sorted(image.getcolors(image.size[0] * image.size[1]), reverse=True)
    curr_color = iter(top_colors)
    while (color := next(curr_color)[1]) == WHITE:
        pass
    return color



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="image-titler",
    version="2.3.0",
    author="The Renegade Coder",
    author_email="jeremy.grifski@therenegadecoder.com",
    description="A image processing utility which provides options for generating thumbnails for various social media platforms.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheRenegadeCoder/image-titler",
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires=">=3.8",
    entry_points={
        "console_scripts": [
            'image-titler = imagetitler.cli:main',
            'image_titler = imagetitler.cli:main',  # For backwards compatibility
        ],
        "gui_scripts": [
            'image-titler-gui = imagetitler.gui:main',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'titlecase',
        'pillow>=6.0.0',
        'pathvalidate',
        'piexif',
        'matplotlib'
    ],
)



"""
The GUI interface for the image-titler script.
"""

import tkinter as tk
import tkinter.ttk as ttk
from pathlib import Path
from tkinter import filedialog
from typing import Optional, List

import pkg_resources
from PIL import ImageTk, Image
from matplotlib import font_manager

from imagetitler.constants import *
from imagetitler.draw import process_images
from imagetitler.parse import parse_input
from imagetitler.store import save_copies

FONTS = {
    f"{f.name} ({f.style}, {f.variant}, {f.weight}, {f.stretch})": f.fname
    for f in font_manager.fontManager.ttflist
}

FILE_TAB_LABEL = "File"
NEW_IMAGE_LABEL = "New Image"
NEW_LOGO_LABEL = "New Logo"
SAVE_AS_LABEL = "Save As"
TITLE_OPTION_LABEL = "Title:"
TIER_OPTION_LABEL = "Tier:"
LOGO_OPTION_LABEL = "Logo:"

COLUMN_WIDTH = 8


class ImageTitlerMain(tk.Tk):
    """
    The main window. This overrides the root class of tk, so we can make a menu.
    The remainder of the GUI is contained within a frame.
    """

    def __init__(self, options):
        super().__init__()
        self.options = options
        self.menu = ImageTitlerMenuBar(self, self.options)
        self.gui = ImageTitlerGUI(self, self.menu, self.options)
        self.gui.pack(anchor=tk.W)

    def update_view(self) -> None:
        """
        Updates what's happening visually in the app.

        :return: None
        """
        self.gui.update_view()

    def save_as(self) -> None:
        """
        A save method which saves our preview. This has to exist because the menu
        has no concept of title. As a result, this method needed to be pulled up
        into main window. That way, we at least decouple the child to parent
        relationship (i.e. children have to concept of siblings, etc.).

        :return: None
        """
        save_copies(
            self.menu.current_edit,
            **self.options
        )


class ImageTitlerGUI(ttk.Frame):
    """
    The main content of the GUI. This contains the preview pane and the option pane.
    """

    def __init__(self, parent, menu, options, **kw):
        super().__init__(parent, **kw)
        self.menu = menu
        self.options = options
        self.logo_path = None
        self.option_pane = ImageTitlerOptionPane(self, self.options)
        self.preview = ImageTitlerPreviewPane(self,
                                              text=f"Select a file using '{FILE_TAB_LABEL}' > '{NEW_IMAGE_LABEL}'")
        self._set_layout()
        self.update_view()

    def update_view(self, *_) -> None:
        """
        Updates this frame visually by controlling what is happening in children components.

        :return: None
        """
        if self.options[KEY_PATH]:
            self._render_preview()
        self._render_logo()

    def _set_layout(self) -> None:
        """
        Sets the layout of the window. Specifically, this function places the option pane
        on the left and the preview pane on the right.

        :return: None
        """
        self.preview.pack(side=tk.RIGHT, expand=tk.YES, fill=tk.BOTH, padx=5, pady=5)
        self.option_pane.pack(side=tk.LEFT, anchor=tk.NW, padx=5, pady=5)

    def _render_preview(self) -> None:
        """
        Renders a preview of the edited image in the child preview pane.

        :return: None
        """
        self.menu.current_edit = process_images(**self.options)
        maxsize = (1028, 1028)
        small_image = self.menu.current_edit[0].copy()
        small_image.thumbnail(maxsize, Image.ANTIALIAS)
        image = ImageTk.PhotoImage(small_image)
        self.preview.config(image=image)
        self.preview.image = image

    def _render_logo(self) -> None:
        """
        Renders a preview of the logo in the options pane.

        :return: None
        """
        logo_path = self.options.get(KEY_LOGO_PATH)
        if logo_path and logo_path != self.logo_path:
            self.logo_path = logo_path
            maxsize = (50, 50)
            small_image = Image.open(logo_path)
            small_image.thumbnail(maxsize, Image.ANTIALIAS)
            image = ImageTk.PhotoImage(small_image)
            self.option_pane.logo_value.config(image=image)
            self.option_pane.logo_value.image = image
            self.option_pane.logo_value.logo_path = self.logo_path
            self.option_pane.logo_state.set(1)


class ImageTitlerPreviewPane(ttk.Label):
    """
    The preview pane is a simple label which contains a preview of the
    image currently being edited.
    """

    def __init__(self, parent, **kw):
        super().__init__(parent, **kw)


class ImageTitlerMenuBar(tk.Menu):
    """
    The menu bar for interactions like loading files and logos.
    """

    def __init__(self, parent: ImageTitlerMain, options: dict):
        super().__init__(parent)
        self.parent: ImageTitlerMain = parent
        self.options: dict = options
        self.current_edit: Optional[List] = None
        self.file_menu: Optional[tk.Menu] = None
        self._init_menu()

    def _init_menu(self) -> None:
        """
        Sets up the menu items.

        :return: None
        """
        menu = tk.Menu(self.parent)
        self.parent.config(menu=menu)

        self.file_menu = tk.Menu(menu, tearoff=0, postcommand=self._save_as_enabled)
        self.file_menu.add_command(label=NEW_IMAGE_LABEL, command=self._new_image)
        self.file_menu.add_command(label=NEW_LOGO_LABEL, command=self._new_logo)
        self.file_menu.add_command(label=SAVE_AS_LABEL, command=self._save_as)
        menu.add_cascade(label=FILE_TAB_LABEL, menu=self.file_menu)

    def _new_image(self) -> None:
        """
        Specifies behavior for when a user selects the "New Image" option.

        :return: None
        """
        self.options[KEY_PATH] = filedialog.askopenfilename(filetypes=FILE_TYPES)
        self.parent.update_view()

    def _new_logo(self) -> None:
        """
        Specifies the behavior when a user selects the "New Logo" option.

        :return: None
        """
        self.options["logo_path"] = filedialog.askopenfilename(filetypes=FILE_TYPES)
        self.parent.update_view()

    def _save_as(self) -> None:
        """
        Specifies the behavior when a user selects the "Save As" option.

        :return: None
        """
        self.options[KEY_OUTPUT_PATH] = filedialog.askdirectory()
        self.parent.save_as()

    def _save_as_enabled(self) -> None:
        """
        Enables/disables the "Save As" option.

        :return: None
        """
        if self.current_edit:
            self.file_menu.entryconfig(2, state=tk.NORMAL)  # TODO: make this not hardcoded
        else:
            self.file_menu.entryconfig(2, state=tk.DISABLED)


class ImageTitlerOptionPane(ttk.Frame):
    """
    The option pane contains a set of options that can be controlled when editing the image.
    Changes are reflected in the preview pane.
    """

    def __init__(self, parent: ImageTitlerGUI, options: dict, **kw):
        super().__init__(parent, **kw)
        self.parent = parent
        self.options = options
        self.title_state: tk.IntVar = tk.IntVar()
        self.title_value: tk.StringVar = tk.StringVar()
        self.tier_state: tk.IntVar = tk.IntVar()
        self.tier_value: tk.StringVar = tk.StringVar()
        self.logo_path: Optional[str] = None
        self.logo_state: tk.IntVar = tk.IntVar()
        self.logo_value: Optional[ttk.Label] = None
        self.font_state: tk.IntVar = tk.IntVar()
        self.font_value: tk.StringVar = tk.StringVar()
        self.size_state: tk.IntVar = tk.IntVar()
        self.size_value: tk.StringVar = tk.StringVar()
        self.rows = list()
        self._init_vars()
        self._init_option_pane()

    def _init_vars(self) -> None:
        """
        Initializes the options pane based on any initial options.

        :return: None
        """
        # TODO: remove this method and add each section to each initialization method
        title = self.options.get(KEY_TITLE)
        ImageTitlerOptionPane._populate_option(title, self.title_value, self.title_state, "")
        tier = self.options.get(KEY_TIER)
        ImageTitlerOptionPane._populate_option(tier, self.tier_value, self.tier_state, list(TIER_MAP.keys())[0])
        font = self.options.get(KEY_FONT)
        self.font_value.set(sorted(list(FONTS.keys()))[0])
        if font != DEFAULT_FONT:
            font = next(k for k, v in FONTS.items() if Path(v).name == font)
            ImageTitlerOptionPane._populate_option(font, self.font_value, self.font_state)
        logo = self.options.get(KEY_LOGO_PATH)
        self.logo_state.set(1 if logo else 0)
        size = self.options.get(KEY_SIZE)
        ImageTitlerOptionPane._populate_option(size, self.size_value, self.size_state, sorted(SIZE_MAP.keys())[0])

    def _init_option_pane(self) -> None:
        """
        Initializes the option pane by generating rows of settings.

        :return: None
        """
        self.rows.append(self._init_title_frame())
        self.rows.append(self.init_font_frame())
        self.rows.append(self._init_size_frame())
        self.rows.append(self._init_tier_frame())
        self.rows.append(self.init_logo_frame())
        for row in self.rows:
            self._layout_option_row(*row[:3])

    def _init_title_frame(self) -> tuple:
        """
        Initializes the row for title information.

        :return: a tuple containing the title container and its two children (see layout_option_row for order)
        """
        title_frame = ttk.Frame(self)
        title_label = ttk.Checkbutton(
            title_frame,
            text=TITLE_OPTION_LABEL,
            variable=self.title_state,
            command=self._update_title,
            width=COLUMN_WIDTH
        )
        title_label.variable = self.title_state
        self.title_value.trace(tk.W, self._update_title)
        title_entry = tk.Entry(title_frame, textvariable=self.title_value)
        return title_frame, title_label, title_entry, "title"

    def _update_title(self, *_) -> None:
        """
        A helper method which serves as the update title functionality.
        This should be triggered when the title is changed.

        :return: None
        """
        if self.title_state.get():
            self.options[KEY_TITLE] = self.title_value.get()
        else:
            self.options[KEY_TITLE] = None
        self.parent.update_view()

    def _init_tier_frame(self) -> tuple:
        """
        Initializes the row for tier information.

        :return: a tuple containing the tier container and its two children (see layout_option_row for order)
        """
        tier_frame = ttk.Frame(self)
        tier_label = ttk.Checkbutton(
            tier_frame,
            text=TIER_OPTION_LABEL,
            variable=self.tier_state,
            command=self._update_tier,
            width=COLUMN_WIDTH
        )
        tier_label.variable = self.tier_state
        tier_option_menu = ttk.Combobox(
            tier_frame,
            textvariable=self.tier_value,
            values=list(TIER_MAP.keys()),
            state="readonly"
        )
        tier_option_menu.bind("<<ComboboxSelected>>", self._update_tier)
        return tier_frame, tier_label, tier_option_menu, KEY_TIER

    def _update_tier(self, *_) -> None:
        """
        A helper method which serves as the update tier functionality.
        This should be triggered when the tier is changed.

        :return: None
        """
        if self.tier_state.get():
            self.options[KEY_TIER] = self.tier_value.get()
        else:
            self.options[KEY_TIER] = None
        self.parent.update_view()

    def init_logo_frame(self) -> tuple:
        """
        Initializes the row for logo information.

        :return: a tuple containing the logo container and its two children (see layout_option_row for order)
        """
        logo_frame = ttk.Frame(self)
        logo_label = ttk.Checkbutton(
            logo_frame,
            text=LOGO_OPTION_LABEL,
            variable=self.logo_state,
            command=self._update_logo,
            width=COLUMN_WIDTH
        )
        logo_label.variable = self.logo_state
        self.logo_value = ttk.Label(logo_frame, text=f"Select a logo using '{FILE_TAB_LABEL}' > '{NEW_LOGO_LABEL}'")
        return logo_frame, logo_label, self.logo_value, KEY_LOGO_PATH

    def _update_logo(self) -> None:
        """
        Renders a preview of the logo in the options pane.

        :return: None
        """
        if self.logo_state.get() and hasattr(self.logo_value, "logo_path"):
            self.options[KEY_LOGO_PATH] = self.logo_value.logo_path
        else:
            self.options[KEY_LOGO_PATH] = None
        self.parent.update_view()

    def init_font_frame(self) -> tuple:
        """
        Initializes the row for font information.

        :return: a tuple containing the font container and its two children (see layout_option_row for order)
        """
        font_frame = ttk.Frame(self)
        font_label = ttk.Checkbutton(
            font_frame,
            text="Font:",
            variable=self.font_state,
            command=self._update_font,
            width=COLUMN_WIDTH
        )
        font_label.variable = self.font_state
        font_list = sorted(FONTS.keys())
        font_menu = ttk.Combobox(
            font_frame,
            textvariable=self.font_value,
            values=font_list,
            state="readonly",
            width=40
        )
        font_menu.bind("<<ComboboxSelected>>", self._update_font)
        return font_frame, font_label, font_menu, KEY_FONT

    def _update_font(self, *_) -> None:
        """
        A helper method which serves as the update font functionality.
        This should be triggered when the font is changed.

        :return: None
        """
        if self.font_state.get():
            self.options[KEY_FONT] = FONTS.get(self.font_value.get())
        else:
            self.options[KEY_FONT] = None
        self.parent.update_view()

    def _init_size_frame(self) -> tuple:
        """
        Initializes the row for size information.

        :return: a tuple containing the size container and its two children (see layout_option_row for order)
        """
        size_frame = ttk.Frame(self)
        size_label = ttk.Checkbutton(
            size_frame,
            text="Size:",
            variable=self.size_state,
            command=self._update_size,
            width=COLUMN_WIDTH
        )
        size_label.variable = self.size_state
        size_menu = ttk.Combobox(
            size_frame,
            textvariable=self.size_value,
            values=sorted(SIZE_MAP.keys()),
            state="readonly"
        )
        size_menu.bind("<<ComboboxSelected>>", self._update_size)
        return size_frame, size_label, size_menu, KEY_SIZE

    def _update_size(self, *_):
        """
        A helper method which serves as the update size functionality.
        This should be triggered when the size is changed.

        :return: None
        """
        if self.size_state.get():
            self.options[KEY_SIZE] = self.size_value.get()
        else:
            self.options[KEY_SIZE] = None
        self.parent.update_view()

    @staticmethod
    def _layout_option_row(frame, label, value) -> None:
        """
        Sets up consistent packing for a row of the option pane.

        :param frame: a row object
        :param label: a row label
        :param value: a row value
        :return: None
        """
        frame.pack(side=tk.TOP, anchor=tk.NW, padx=10, pady=5, expand=tk.YES, fill=tk.X)
        label.pack(side=tk.LEFT)
        value.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X)

    @staticmethod
    def _populate_option(option: str, value: tk.StringVar, state: tk.IntVar, default_option: str = None):
        if option:
            value.set(option)
            state.set(1)
        else:
            value.set(default_option)
            state.set(0)


def main():
    """
    The GUI main function.

    :return: None
    """
    options: dict = vars(parse_input())
    root = ImageTitlerMain(options)
    version = pkg_resources.require("image-titler")[0].version
    root.title(f"The Renegade Coder Image Titler {version}")
    root.iconphoto(False, tk.PhotoImage(file=TRC_ICON))
    root.mainloop()


if __name__ == '__main__':
    main()
